#!/usr/bin/env python
# encoding:utf8
#
# This source file is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This source file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Date: 2020
#
# Authors: Moritz Layer (m.layer@fz-juelich.de) [1,2]
#
# Affiliations:
# [1] Institute of Neuroscience and Medicine (INM-6) and Institute for Advanced
#     Simulation (IAS-6) and JARA Institut Brain Structure-Function
#     Relationships (INM-10), Jülich Research Centre, Jülich, Germany
# [2] RWTH Aachen University, Aachen, Germany

'''
Create plot for simulation part of DisCo manuscript.

Usage: plot_simulation_composition.py

Options:
    -h, --help                   show this information

'''

import docopt
import yaml
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from svgutils.compose import (Figure, Panel, SVG)

from utils_simulation import parameter_string, mean_distribution


def mm2inch(x):
    return x / 25.4


def labels(ax, fontsize=16, title=None, xlabel=None, ylabel=None, zlabel=None,
           legend=True, color='k', xcolor='k', ycolor='k', zcolor='k',
           suptitle=None, fig=None, legendfontsize=None, mfs=None):
    """ To add a suptitle provide also a figure handle 'fig'. """
    if title:
        ax.set_title(title, fontsize=fontsize, color=color)

    if suptitle:
        if fig:
            fig.suptitle(suptitle, fontsize=fontsize + 1, color=color)
        else:
            print('Provide a figure handle to add a suptitle')

    if xlabel:
        ax.set_xlabel(xlabel, fontsize=fontsize, color=xcolor)

    if ylabel:
        ax.set_ylabel(ylabel, fontsize=fontsize, color=ycolor)

    if zlabel:
        ax.set_zlabel(zlabel, fontsize=fontsize, color=zcolor)

    if legend:
        ax.legend(loc='best', handlelength=1)

    if legendfontsize:
        ax.legend(loc='best', handlelength=1, fontsize=legendfontsize)

    if xcolor == ycolor:
        ax.tick_params('both', labelsize=fontsize - 1, color=xcolor)
    else:
        ax.tick_params('x', labelsize=fontsize - 1, colors=xcolor)
        ax.tick_params('y', labelsize=fontsize - 1, colors=ycolor)
    if mfs:
        print('OK but this does not work => fontsize has no effect here and '
              'above')
        ax.tick_params('x', labelsize=7, colors=xcolor)


def distribution(ax, data, bins=None, histtype='bar', color='darkseagreen',
                 lw=1., linestyle='-', normed=False, label=None, fontsize=11):
    """
    Plots a histogram of data in given axes.

    If data is a multidimensional array, flattens it and plots the distribution
    of all (N) numbers.
    If bins are not given, divides the distribution into floor(sqrt(N)) bins.
    """
    if isinstance(data, np.ndarray):
        if len(data.shape) > 1:
            data = data.flatten()

    ids = np.where([not np.isnan(d) for d in data])[0]

    if bins is None:
        bins = int(np.floor(np.sqrt(np.size(data))))
        print('size:', np.size(data), 'bins:', bins)

    if label:
        n, bins, _ = ax.hist(data[ids], bins=bins, histtype=histtype, lw=lw,
                             linestyle=linestyle, color=color, density=normed,
                             label=label)
    else:
        n, bins, _ = ax.hist(data[ids], bins=bins, histtype=histtype, lw=lw,
                             linestyle=linestyle, color=color, density=normed)

    ax.set_xlim([min(bins), max(bins)])
    ax.set_ylim([0, max(n) * 1.2])

    return n, bins


def corrcoef(a, b):
    a = a.flatten()
    b = b.flatten()
    return np.corrcoef(a, b)[0][1]


def histograms(ax, data0, data1, n_bins):
    """ Plots histogram of data0 and data1. """
    data = np.array([data0, data1])
    limits = [np.min(data), np.max(data)]
    bins = np.linspace(limits[0], limits[1], n_bins)

    distribution(ax, data[0], bins=bins, label='N1',
                 color='steelblue', histtype='step')
    distribution(ax, data[1], bins=bins, label='N2',
                 color='cyan', histtype='step')

    asp = np.diff(ax.get_xlim())[0] / np.diff(ax.get_ylim())[0]
    ax.set_aspect(asp)


def scatter_plot(ax, data0, data1, ticks=None, fontsize=11, limits=None,
                 markersize=3, samplesize=None):
    """ Plots scatter plot of data0 and data1. """
    cc = np.corrcoef(data0, data1)[0, 1]

    if samplesize:
        indices = np.arange(len(data0))
        sample_indices = np.random.choice(indices, size=samplesize,
                                          replace=False)
        data0 = data0[sample_indices]
        data1 = data1[sample_indices]

    data = np.array([data0, data1])
    if not limits:
        limits = [np.min(data), np.max(data)]
    range = limits[1] - limits[0]

    ax.plot(data0, data1, 'ok', ms=markersize, markerfacecolor='None')
    # equiline
    ax.plot(limits, limits, color='grey', linewidth=1.5)
    ax.text(limits[0] + 0.04 * range, limits[1] - 0.03 * range,
            r'$\rho$=%.2f' % cc, ha='left',
            va='top', fontsize=fontsize - 2)
    ax.set_xlim(limits)
    ax.set_ylim(limits)

    asp = np.diff(ax.get_xlim())[0] / np.diff(ax.get_ylim())[0]
    ax.set_aspect(asp)
    asp = np.diff(ax.get_xlim())[0] / np.diff(ax.get_ylim())[0]
    ax.set_aspect(asp)

    if ticks:
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)

    asp = np.diff(ax.get_xlim())[0] / np.diff(ax.get_ylim())[0]
    ax.set_aspect(asp)


def get_simulation_data(parameters, rates_path):
    """ Loading routine for simulated data. """
    temp_dict = parameters.copy()
    temp_dict.pop('chunks')
    output_string = parameter_string(temp_dict)
    rates = np.load('{}{}.npy'.format(rates_path, output_string))
    return rates


def corr_coef_within(rates_half):
    """ Calculates corr coef between first and second half of simulation. """
    corr_coef_within = np.zeros(2)
    corr_coef_within[0] = np.corrcoef(rates_half[:2])[0, 1]
    corr_coef_within[1] = np.corrcoef(rates_half[2:])[0, 1]
    return corr_coef_within


def corr_coef_between(rates_half):
    """ Calculates corr coef between two halves of twp simulation. """
    corr_coef_between = np.zeros(4)
    corr_coef_between[0] = np.corrcoef([rates_half[0], rates_half[2]])[0, 1]
    corr_coef_between[1] = np.corrcoef([rates_half[0], rates_half[3]])[0, 1]
    corr_coef_between[2] = np.corrcoef([rates_half[1], rates_half[2]])[0, 1]
    corr_coef_between[3] = np.corrcoef([rates_half[1], rates_half[3]])[0, 1]
    return corr_coef_between


def get_half_data(parameters0, parameters1, rate_path_half0, rate_path_half1):
    """ Loading routine for data of halves of simulation. """
    rates_half = []
    rates_half.append(get_simulation_data(parameters0, rate_path_half0))
    rates_half.append(get_simulation_data(parameters0, rate_path_half1))
    rates_half.append(get_simulation_data(parameters1, rate_path_half0))
    rates_half.append(get_simulation_data(parameters1, rate_path_half1))
    return rates_half


def within_between_comparison(ax, parameters0, parameters1, rate_path_half0,
                              rate_path_half1, cov_path_half0, cov_path_half1,
                              fontsize=11):
    """ Plot corr coefs within and between halves of two simulations. """
    rates_half = get_half_data(parameters0, parameters1,
                               rate_path_half0, rate_path_half1)
    covs_half = get_half_data(parameters0, parameters1,
                              cov_path_half0, cov_path_half1)
    corr_rates_within = corr_coef_within(rates_half)
    corr_rates_between = corr_coef_between(rates_half)
    corr_covs_within = corr_coef_within(covs_half)
    corr_covs_between = corr_coef_between(covs_half)

    color_within = 'grey'
    color_between = 'black'
    markersize = 5

    ax.plot(np.zeros(len(corr_rates_within)), corr_rates_within,
            'ok', color=color_within, ms=markersize, markerfacecolor='None',
            label='w')
    ax.plot(np.zeros(len(corr_rates_between)), corr_rates_between,
            'ok', color=color_between, ms=markersize, markerfacecolor='None',
            label='b')
    ax.plot(np.ones(len(corr_covs_within)) * 1.3, corr_covs_within,
            'ok', color=color_within, ms=markersize, markerfacecolor='None')
    ax.plot(np.ones(len(corr_covs_between)) * 1.3, corr_covs_between,
            'ok', color=color_between, ms=markersize, markerfacecolor='None')

    ax.set_xlim([-0.5, 2])

    xticks = [0, 1.3]
    xticklabels = ['rates', 'covs']
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticklabels)

    ax.set_ylim([0, 1.1])

    asp = np.diff(ax.get_xlim())[0] / np.diff(ax.get_ylim())[0]
    ax.set_aspect(asp)


def get_data_for_corr_coef_of_r(rs, rates_path, config_path):
    """ Loads data for calculating corr coef as function of r. """
    mean_rates_r = []
    for r in rs:
        id = str(r).replace('.', '')
        for i in range(2):
            parameters_yaml = config_path + 'params_r{}_{}.yaml'.format(id, i)
            with open(parameters_yaml, 'r') as stream:
                parameters = yaml.safe_load(stream)
            mean_rates_r.append(get_simulation_data(parameters, rates_path))
    return mean_rates_r


def corr_coefs_of_r(ax, rates_path, covs_path, config_path, rs, fontsize=11):
    """ Calculate correlation coefficient as function of r. """
    mean_rates_r = get_data_for_corr_coef_of_r(rs, rates_path, config_path)
    cross_covs_r = get_data_for_corr_coef_of_r(rs, covs_path, config_path)

    corr_coef_mean_rates = np.zeros(len(rs))
    corr_coef_cross_cov = np.zeros(len(rs))
    for i in range(len(rs)):
        corr_coef_mean_rates[i] = corrcoef(mean_rates_r[2 * i],
                                           mean_rates_r[2 * i + 1])
        corr_coef_cross_cov[i] = corrcoef(cross_covs_r[2 * i],
                                          cross_covs_r[2 * i + 1])

    ax.plot(rs, corr_coef_mean_rates, label='rate', color='grey')
    ax.plot(rs, corr_coef_cross_cov, label='cov', color='black')
    xticks = [0.2, 0.5, 0.8]
    yticks = [0, 0.3, 0.6, 0.9]
    ax.set_ylim([-0.05, 0.95])
    ax.set_xticks(xticks)
    ax.set_yticks(yticks)
    asp = np.diff(ax.get_xlim())[0] / np.diff(ax.get_ylim())[0]
    ax.set_aspect(asp)


def simulation_plot(parameters0, parameters1, rate_path_half0, rate_path_half1,
                    cov_path_half0, cov_path_half1, rates_path, covs_path,
                    config_path, rs):
    """ Plotting routine for simulation figure in manuscript. """
    fontsize = 11
    legendfontsize = fontsize - 2

    figsize = (mm2inch(180), mm2inch(180) * 0.75)
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    # fig = plt.figure(figsize=figsize)
    # plt.tight_layout()
    gs = gridspec.GridSpec(3, 4, figure=fig)
    ax0 = fig.add_subplot(gs[0, 0])
    ax1 = fig.add_subplot(gs[0, 2])
    ax2 = fig.add_subplot(gs[0, 3])
    ax3 = fig.add_subplot(gs[1, 2])
    ax4 = fig.add_subplot(gs[1, 3])
    ax5 = fig.add_subplot(gs[2, 0])
    ax6 = fig.add_subplot(gs[2, 1])
    ax7 = fig.add_subplot(gs[2, 2])
    ax8 = fig.add_subplot(gs[2, 3])

    # plot whitespace
    ax0.set_axis_off()
    asp = np.diff(ax0.get_xlim())[0] / np.diff(ax0.get_ylim())[0]
    ax0.set_aspect(asp)

    # get mean external input for both networks
    mean_input0 = mean_distribution(parameters0)
    mean_input1 = mean_distribution(parameters1)

    histograms(ax1, mean_input0, mean_input1, n_bins=40)
    labels(ax1, ylabel='pdf', xlabel='mean input', fontsize=fontsize,
           legendfontsize=legendfontsize)
    ax1.legend(borderpad=0.2, handlelength=0.4, handletextpad=0.2,
               loc='upper right', bbox_to_anchor=(1.05, 1.05))
    ax1.set_xticklabels([5, 7.5, 10, 12.5, 15])
    ax1.set_yticks([0, 80, 160])
    ax1.set_ylabel('pdf', labelpad=-2)

    ticks = [6, 8, 10, 12]
    scatter_plot(ax2, mean_input0, mean_input1, ticks, markersize=0.1,
                 samplesize=1500)
    labels(ax2, xlabel='mean input', ylabel='mean input', legend=False,
           fontsize=fontsize)

    within_between_comparison(ax3, parameters0, parameters1, rate_path_half0,
                              rate_path_half1, cov_path_half0, cov_path_half1)
    labels(ax3, ylabel='corr coef', legend=True, legendfontsize=legendfontsize,
           fontsize=fontsize)
    ax3.legend(handlelength=0.8, handletextpad=0.1, borderpad=0.3,
               loc='lower left', bbox_to_anchor=(-0.06, -   0.06))

    corr_coefs_of_r(ax4, rates_path, covs_path, config_path, rs)
    labels(ax4, xlabel='$R$', ylabel='corr coef', legend=True,
           legendfontsize=legendfontsize, fontsize=fontsize)
    ax4.legend(handlelength=0.8, handletextpad=0.3, loc='lower left',
               bbox_to_anchor=(-0.06, -0.06))

    # load data from full simulation for histograms
    rates = []
    rates.append(get_simulation_data(parameters0, rates_path))
    rates.append(get_simulation_data(parameters1, rates_path))
    cross_covs = []
    cross_covs.append(get_simulation_data(parameters0, covs_path))
    cross_covs.append(get_simulation_data(parameters1, covs_path))

    # remove silent neurons as they would not be visible in experiments
    rates = np.array(rates)
    active_indices = np.where(np.logical_and(rates[0] > 1e-3,
                                             rates[1] > 1e-3))[0]
    rates_active = [rates[active_indices] for rates in rates]

    # create list of neuron index pairs for covariances (compare analysis.py)
    index_pairs = []
    N = len(rates[0])
    for i in range(N):
        for j in range(i + 1, N):
            index_pairs.append((i, j))
    # find active pairs
    active_pair_indices = []
    for id, (i, j) in enumerate(index_pairs):
        if i in active_indices and j in active_indices:
            active_pair_indices.append(id)
    # remove silent pairs
    cross_covs = np.array(cross_covs)
    cross_covs_active = [cross_covs[active_pair_indices]
                         for cross_covs in cross_covs]

    histograms(ax5, rates_active[0], rates_active[1], n_bins=10)
    labels(ax5, ylabel='pdf', xlabel='firing rate', fontsize=fontsize,
           legendfontsize=legendfontsize)
    ax5.legend(handlelength=0.5, handletextpad=0.2, loc='upper right',
               bbox_to_anchor=(1.05, 1.05))
    ax5.set_yticks([0, 15, 30])

    limits = [-0.2, 4.2]
    ticks = [0, 2, 4]
    scatter_plot(ax6, rates[0], rates[1], ticks=ticks, markersize=0.7,
                 limits=limits)
    labels(ax6, xlabel='firing rate', ylabel='firing rate', fontsize=fontsize,
           legend=False)

    ax7.set_ylim([0, 600])
    histograms(ax7, cross_covs_active[0], cross_covs_active[1], n_bins=50)
    labels(ax7, ylabel='pdf', xlabel='cov', fontsize=fontsize,
           legendfontsize=legendfontsize)
    ax7.legend(borderpad=0.3, handlelength=0.5, handletextpad=0.2,
               loc='upper left', bbox_to_anchor=(-0.06, 1.06))
    ax7.set_yticks([0, 500])
    ax7.set_ylabel('pdf', labelpad=-8)
    ax7.set_xticklabels([-0.02, -0.01, 0, 0.01, 0.02])

    limits = [-0.015, 0.015]
    scatter_plot(ax8, cross_covs[0], cross_covs[1], limits=limits,
                 markersize=0.1, samplesize=5000)
    labels(ax8, xlabel='cov', ylabel='cov', fontsize=fontsize, legend=False)
    ticks = [-0.01, 0, 0.01]
    ax8.set_xticks(ticks)
    ax8.set_yticks(ticks)
    ax8.set_xticklabels(ticks)
    ax8.set_yticklabels(ticks)
    ax8.set_ylabel('cov', labelpad=-5)

    # add labels to panels
    axs = [ax0, ax1, ax2, ax3, ax4, ax5, ax6, ax7, ax8]
    label = ['A', 'B1', 'B2', 'C', 'D', 'E1', 'E2', 'F1', 'F2']

    for n, ax in enumerate(axs):
        ax.text(-0.1, 1.1, label[n], transform=ax.transAxes,
                size=11, weight='bold')

    output_filename = 'fig5.svg'
    plt.savefig(output_filename)
    plt.close()

    # combine matplotlib figure with svg sketch
    Figure("14cm", "10.5cm",
           Panel(SVG(output_filename)).move(10, 10),
           Panel(SVG("fig5a.svg").move(8, 10).scale(2.1)),
           ).save(output_filename)


if __name__ == '__main__':

    config_path = 'config/'
    parameters_yamls = [config_path + 'params_r08_0.yaml',
                        config_path + 'params_r08_1.yaml']

    param_list = []
    for parameters_yaml in parameters_yamls:
        with open(parameters_yaml, 'r') as stream:
            param_list.append(yaml.safe_load(stream))

    data_path = 'data/nest/'

    rate_path_half0 = data_path + 'mean_rates_half_1/'
    rate_path_half1 = data_path + 'mean_rates_half_2/'
    cov_path_half0 = data_path + 'cross_cov_half_1/'
    cov_path_half1 = data_path + 'cross_cov_half_2/'

    rates_path = data_path + 'mean_rates/'
    covs_path = data_path + 'cross_cov_output/'

    rs = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])

    simulation_plot(param_list[0], param_list[1],
                    rate_path_half0, rate_path_half1,
                    cov_path_half0, cov_path_half1,
                    rates_path, covs_path, config_path, rs)
