# encoding:utf8
#
# This source file is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This source file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random
import numpy as np
from scipy import signal


def get_sample_indices(N, sample_size, sample_seed):
    random.seed(sample_seed)
    indices = list(range(N))
    sample_indices = random.sample(indices, sample_size)
    return sample_indices


def weight(r, N, p):
    """ inhibitory weights at given spectral radius with N/2 active neurons """
    return - r / np.sqrt(N / 2 * p * (1 - p))


def mean_distribution(parameters):
    N = int(parameters['N'])
    np_seed = int(parameters['np_seed'])
    np_seed_extra = int(parameters['np_seed_extra'])
    mean_mean_input = float(parameters['mean_mean_input'])
    std_mean_input = float(parameters['std_mean_input'])
    epsilon = float(parameters['epsilon'])
    # set numpy seed
    np.random.seed(np_seed)
    # draw mean noise values for neurons from normal distribution
    mean_input = np.random.normal(mean_mean_input,
                                  np.sqrt((1 - epsilon) * std_mean_input**2),
                                  size=N)
    # set numpy seed for extra
    np.random.seed(np_seed_extra)
    little_extra_to_mean_input = np.random.normal(
        0, np.sqrt(epsilon * std_mean_input**2), size=N)
    mu = mean_input + little_extra_to_mean_input
    return mu


def connectivity_string(parameters):
    N = int(parameters['N'])
    p = float(parameters['p'])
    r = float(parameters['r'])
    return 'N{}_p{}_r{}'.format(N, p, r)


def parameter_string(parameters):
    """ Create a parameters string from a dictionary """
    temp_dict = parameters.copy()
    parameter_string = ''
    for key in sorted(list(temp_dict.keys())):
        parameter_string += str(key) + str(temp_dict[key]) + '_'
    parameter_string = parameter_string[:-1]
    return parameter_string


def parameter_string_without_analysis_parameters(parameters):
    temp_dict = parameters.copy()
    temp_dict.pop('tau_max')
    temp_dict.pop('sample_size')
    temp_dict.pop('sample_seed')
    temp_dict.pop('chunks')
    output_string = parameter_string(temp_dict)
    return output_string


def cross_corr(rate1, rate2, bin_width, T, tau_max):
    '''
    Calculate crosscorrelation of given firing rates.

    NOTE: T needs to be at least 2*tau_max.

    Parameters:
    -----------
    rate1: np.ndarray
        First rate in Hz.
    rate2: np.ndarray
        Second rate in Hz.
    bin_width: float
        Bin width in ms.
    T: int
        Total simulation time in ms (without initalisation time).
    tau_max: float
        Maximal time lag considered in ms.

    Returns:
    --------
    time: np.ndarray
        Array from -tau_max to tau_max in bin_width steps.
    autof: np.ndarray
        Array of crosscorrelations as function of time.
    '''

    crossf = (signal.fftconvolve(rate1, rate2[::-1], mode='same')
              * bin_width / T)
    mid = int(len(crossf) / 2)
    crossf = crossf[mid - int(tau_max / bin_width):
                    mid + int(tau_max / bin_width)]
    time = np.arange(-tau_max, tau_max, bin_width)
    return time[1:], crossf[1:]
